/* 
*   GENFONT: Convert font segment file to font data.  by  John H. DuBois III.
*   87/12/27 jhdiii@ucscb.ucsc.edu
*   91/05/04 johnd@sco.com Ported to SYSV
*   92/09/14 jhdiii@armory.com Changed to accept any char value
*   
*   Input:
*   Font segment file is a file containing fonts for some subset of chars
*   0-127 in a 5x5 segment array form.  The file should begin with a template
*   constisting of a 5x5 character array where each char in the array is the
*   one that must be specified at that position if that segment is to be part
*   of the char; this template is used for error checking.
*   Each font specification should consist of the character number on a line,
*   followed by a 5x5 character array where each character given indicates that
*   that segment should be part of the char being specified.  The char given
*   in each position must match the char in that position in the template.  If
*   the segment is not to be part of the char, a space should be put in that
*   position.  Trailing spaces need not be given.
*   Output:
*   128 long ints are written to stdout; one for each char, proceeding from
*   char 0 to char 127.  The 25 low bits in each int represent the font
*   specification for that char.  A 1 bit indicates that a segment should be
*   on; a 0 bit that it should be off.  Bit 0 is for the top right segment,
*   bit 24 is for the bottom left segment; segment specs proceed from right
*   to left and top to bottom.
*   The ints are written in hex in C array initializer format (separated by
*   commas), with 8 ints per line.
*   Following this, 128 integers are written in the same format.  The four
*   nybbles of each int, from high nybble to low nybble, represent the top and
*   bottom row and left and right column occupied by the character respectively;
*   this information is used for proportional spacing when printing the
*   characters.  Each number is between 0 and 4; 0 is the leftmost row or top
*   column, 4 is the rightmost row or bottom column.
*   An empty character (no segments used; either a space or an unspecified char)
*   is taken to occupy the center row and center column.
*/
#include <stdio.h>
#include <stdlib.h>
#include "segfont.h"

#ifdef __MSDOS__
#include <process.h>
#endif

#ifdef _M_SYSV
#define M_SYS5
#endif

#ifdef M_SYS5
#include <string.h>
void GenFont(FILE *InFile,long Font[]);
int CheckLine(char Template[],char Line[]);
void ReadLine(FILE *InFile,char *String,int Num);
long ConvertFont(char FontSpec[][5]);
void WriteFont(long Font[]);
void FindDim(long Font[],int Dim[]);
int RowLimits(long FontData);
int ColLimits(long FontData);
void WriteDim(int Dim[]);
#else
#include <strings.h>
void GenFont();
int CheckLine();
void ReadLine();
long ConvertFont();
void WriteFont();
void FindDim();
int RowLimits();
int ColLimits();
void WriteDim();
#endif

#define TRUE 1
#define FALSE 0
#define ExTest(Test,Mess) if (Test) {fputs(Mess,stderr); exit(2);}

main(int argc,char *argv[])
{
    FILE *InFile;
    static long Font[MAX_CHAR + 1];
    int Dim[MAX_CHAR + 1];	/* Dimensions of chars */

    if (argc > 1) {
	if ((InFile = fopen(argv[1],"r")) == NULL) {
	    puts("Error opening file.");
	    exit(2);
	}
    }
    else
	InFile = stdin;
    GenFont(InFile,Font);	/* Generate font array */
    WriteFont(Font);		/* Write font array in hex to stdout */
    putchar('\n');
    /* Find dimensions of chars (first & last row & col used) */
    FindDim(Font,Dim);
    WriteDim(Dim);		/* Write dimensions in hex to stdout */
}

/* 
*   Find first & last row & col of each font spec in Font and pack into an
*   integer array. 
*/
void FindDim(long Font[],int Dim[])
{
    int I;
    long FontData;

    for (I = 0; I <= MAX_CHAR; I++) {
	FontData = Font[I];
	if (!FontData)
	    Dim[I] = 0x2222;	/* If font is empty, set all dims to 2. */
	else
	    Dim[I] = (RowLimits(FontData) << 8) | ColLimits(FontData);
    }
}

/* 
*   Find first and last row occupied by FontData 
*/
int RowLimits(long FontData)
{
    int Row,First = -1,Last;

    for (Row = 0; Row < 5; Row++) {
	if (FontData & 0x1fL) {	/* Check low 5 bits */
	    Last = Row;		/* Update last occupied row */
	    if (First == -1)	/* If first occupied row found, set First */
		First = Row;
	}
	FontData >>= 5;		/* Shift next row bits to low 5 bits */
    }
    return((First << 4) | Last);
}

/* 
*   Find first and last column occupied in FontData 
*/
int ColLimits(long FontData)
{
    int Col,First = -1,Last;

    for (Col = 0; Col < 5; Col++) {
	if (FontData & 0x1084210L) {
	 /* If low bit (leftmost col) of any row set */
	    Last = Col;		/* Update last col */
	    if (First == -1)	/* If first occupied col found, set First */
		First = Col;
	}
	FontData <<= 1; /* Shift next column bits into the low bit of each row */
    }
    return((First << 4) | Last);
}

/* 
*   Writes the values in Dim to stdout in C array initializer form 
*/
void WriteDim(int Dim[])
{
    int I;

    for (I = 0; I <= MAX_CHAR; I++) {
	/* Print in hex, preceded by 0x, with 4 hex digits per int and
	    padded on the left with 0's. */
	printf("0x%04x,",Dim[I]);
	if (I % 11 == 10)		/* Print newline after every 11 ints */
	    putchar('\n');
    }
}

/* 
*   Writes the values in Font to stdout in C array initializer form 
*/
void WriteFont(long Font[])
{
    int I;

    for (I = 0; I <= MAX_CHAR; I++) {
	/* Print in hex, preceded by 0x, with at least 6 hex digits per int and
	    padded on the left with 0's. */
	printf("0x%06lxL,",Font[I]);
	if (I % 7 == 6)		/* Print newline after every 7 ints */
	    putchar('\n');
    }
}

/* 
*   Reads InFile and converts data in it to Font 
*/
void GenFont(FILE *InFile,long Font[])
{
    int I;
    char Template[5][5];	/* Char that must be used for each segment */
    char FontSpec[5][5];	/* Font to be converted */
    char InLine[80];	/* Line read in */
    int CharNum,	/* ASCII number of char that font is given for */
    LineNum,		/* Line number of font file being acted on */
    ColNum;		/* Column number of font spec that error occured on */

    for (I = 0; I < 5; I++)	/* Read template from first 5 lines of file */
	ReadLine(InFile,Template[I],5);
    LineNum = 6;
    while (!feof(InFile)) {
	do {
	    fgets(InLine,80,InFile);	/* Skip blank lines */
	    LineNum++;
	} while (!feof(InFile) && InLine[0] == '\0');
	if (!feof(InFile)) {
	    CharNum = atoi(InLine);	/* Read character number given */
	    if (!(0 < CharNum && CharNum <= MAX_CHAR)) {
		fprintf(stderr,"Bad character number at line %d: %d. (exiting)\n",
		LineNum - 1,CharNum);
		exit(2);
	    }
	    for (I = 0; I < 5; I++) {		/* Read font */
		ReadLine(InFile,FontSpec[I],5);
		LineNum++;
		ExTest(feof(InFile),"Unexpected end of file.\n");
		/* Compare font given to template */
		ColNum = CheckLine(Template[I],FontSpec[I]);
		if (ColNum)
		    fprintf(stderr,"Bad char @ line %d; col %d, row %d of "
		     "char font %d is \'%c\'; must be ' ' or \'%c\'.\n",
		     LineNum - 1,ColNum,I + 1,CharNum,FontSpec[I][ColNum - 1],
		     Template[I][ColNum - 1]);
	    }
	    Font[CharNum] = ConvertFont(FontSpec);/* Convert font to long int */
	}
    }
}

/* 
*   Converts an array of 5 x 5 chars to a long int.  If a character in the
*   array is a space, the corresponding bit in the int is set to 0; if
*   nonspace, the bit is set to 1.  The bits that correspond to each char
*   location are as follows:
*   			 4  3  2  1  0
*   			 9  8  7  6  5
*   			14 13 12 11 10
*   			19 18 17 16 15
*   			24 23 22 21 20
*   The high bits are set to 0. 
*/
long ConvertFont(char FontSpec[][5])
{
    long Temp;
    int Row,Col;

    Temp = 0L;
    for (Row = 4; Row >= 0; Row--)
	for (Col = 0; Col < 5; Col++) {
	    Temp <<= 1;
	    if (FontSpec[Row][Col] != ' ')
		Temp |= 1L;
	}
    return(Temp);
}

/* 
*   Compares the characters in Lines to the characters in Template.  If any
*   nonspace char is given in Line, it must be the same as the char in the
*   same position in Template.  Template and Line are arrays of 5 chars.
*   If any bad chars are given, the char position (1-5) is returned;
*   otherwise FALSE is returned.
*/
int CheckLine(char Template[],char Line[])
{
    int I;

    for (I = 0; I < 5; I++)
	if (Line[I] != ' ' && Line[I] != Template[I])
	    return(I + 1);
    return(FALSE);
}

/* 
*   Reads a line from InFile of up to 80 chars; assigns the leftmost Num chars
*   to String.  If the line read in is less than Num chars long, it is padded
*   on the right with blanks.
*/
void ReadLine(FILE *InFile,char *String,int Num)
{
    char InString[80];
    int I;

    fgets(InString,80,InFile);
    strncpy(String,InString,Num);
    for (I = strcspn(String,"\n"); I < Num; I++)
	String[I] = ' ';
}
